import path from 'path';
import 'query-string';
import fs from 'fs-extra';
import postcss from 'postcss';
import colorConverter from 'postcss-color-converter';
import resolveAsync, { sync } from 'resolve';
import * as sfcparser from '@vue/compiler-sfc';
import * as htmlparser from 'htmlparser2';
import * as jsparser from '@babel/parser';
import traverse from '@babel/traverse';
import { createFilter } from '@rollup/pluginutils';
import { inspect } from 'util';
import purgecss from '@fullhuman/postcss-purgecss';

function normalizePath(...paths) {
  const f = path.join(...paths).replace(/\\/g, "/");
  if (/^\.[/\\]/.test(paths[0])) return `./${f}`;
  return f;
}
const relativePath = (from, to) => normalizePath(path.relative(from, to));
const humanlizePath = file => relativePath(process.cwd(), file);
const resolveId = async (id, options = {}) => new Promise(resolve => resolveAsync(id, options, (err, res) => err ? resolve() : resolve(res)));
/** Every project is expected to have a main file which imports both vue3-ui css packages */
const isMain = code => new RegExp("import.*@bulvar/bulma/css/bulma.css").test(code) && new RegExp("import.*@pathscale/bulma-extensions").test(code);
const getDeclaredVariables = code => {
  var _code$match;
  return (_code$match = code.match(/(?<=--)(.*?)(?=:)/g)) !== null && _code$match !== void 0 ? _code$match : [];
};
const isVariableUsed = (v, code) => code.includes(`var(--${v})`);
/** The imports are nuked and replaced with an import for a fake file that will be dynamically created by the plugin */
const injectFakeBundle = code => `import './vue3-ui-bundle.css';
  ${code.replace(/import.*@bulvar\/bulma\/css\/bulma\.css.*/gi, "").replace(/import.*@pathscale\/bulma-extensions.*/gi, "").replace(/import.*user\.css.*/gi, "")}`;
const makeVue3UiBundle = async id => {
  const bulmaCSSFile = await resolveId("@bulvar/bulma/css/bulma.css");
  if (!bulmaCSSFile) throw new Error("TRANSFORM - BULMA CSS NOT FOUND");
  const extensionsCSSFile = await resolveId("@pathscale/bulma-extensions-css-var");
  if (!extensionsCSSFile) throw new Error("TRANSFORM - EXTENSIONS CSS NOT FOUND");
  let bulmaCSS = await fs.readFile(bulmaCSSFile, "utf-8");
  const extensionsCSS = await fs.readFile(extensionsCSSFile, "utf-8");
  const userCSSFile = normalizePath(path.dirname(id), "user.css");
  if (!(await fs.pathExists(userCSSFile))) return bulmaCSS + extensionsCSS;
  let userCSS = await fs.readFile(userCSSFile, "utf-8");
  console.log(`TRANSFORM - USER CSS FOUND (${humanlizePath(userCSSFile)})`);
  // maps any color declaration to --xxx: hsl(a, b, c)
  userCSS = postcss(colorConverter({
    outputColorFormat: "hsl"
  })).process(userCSS).css;
  // nuke comments
  userCSS = userCSS.replace(/\/\*[^*]*\*+([^*/][^*]*\*+)*\//gi, "");
  // map hsl(a, b, c) into for variable declarations, each with postfix -h -s -l -a
  userCSS = userCSS.replace(/--(.*):\shsl\((\d*),\s*(\d*)%,\s*(\d*)%\);?/gi, (_, name, h, s, l) => `${_.endsWith(";") ? _ : _.slice(0, -1)}
        --${name}-h: ${h};
        --${name}-s: ${s}%;
        --${name}-l: ${l}%;
        --${name}-a: 1;`);
  // console.log(`TRANSFORM - USER CSS RESULT:\n`, userCSS);
  // replaces on bulmaCSS the definition of the variables swaping line for line
  const vars = getDeclaredVariables(userCSS);
  for (const v of vars) {
    var _decl$exec$, _decl$exec;
    const decl = new RegExp(`.*${v}:.*`);
    bulmaCSS = bulmaCSS.replace(decl, (_decl$exec$ = (_decl$exec = decl.exec(userCSS)) === null || _decl$exec === void 0 ? void 0 : _decl$exec[0]) !== null && _decl$exec$ !== void 0 ? _decl$exec$ : "");
  }
  return bulmaCSS + extensionsCSS;
};
const kebabCase = s => {
  const str = s.slice(1).replace(/[A-Z]/g, m => `-${m.toLowerCase()}`);
  return s.charAt(0).toLowerCase() + str;
};
const camelCase = s => {
  const str = s.slice(1).replace(/-([a-z])/g, (_, $1) => $1.toUpperCase());
  return s.charAt(0).toLowerCase() + str;
};
const pascalCase = s => {
  const str = s.slice(1).replace(/-([a-z])/g, (_, $1) => $1.toUpperCase());
  return s.charAt(0).toUpperCase() + str;
};

function parseSFC(code, id) {
  const {
    descriptor: {
      template,
      script
    }
  } = sfcparser.parse(code, {
    sourceMap: false,
    sourceRoot: process.cwd(),
    filename: id,
    pad: "line"
  });
  return {
    template,
    script
  };
}
function isVueSFC(id) {
  return /\.vue$/i.test(id);
}

const parserOpts = {
  sourceType: "unambiguous",
  plugins: ["asyncGenerators", "bigInt", "classPrivateMethods", "classPrivateProperties", "classProperties", "decimal", "doExpressions", "dynamicImport", "exportDefaultFrom", "functionBind", "functionSent", "importMeta", "jsx", "logicalAssignment", "nullishCoalescingOperator", "numericSeparator", "objectRestSpread", "optionalCatchBinding", "optionalChaining", "partialApplication", "placeholders", "privateIn", "throwExpressions", "typescript"]
};
const vue3ui = sync("@pathscale/vue3-ui", {
  basedir: __dirname,
  packageFilter(pkg) {
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
    if (pkg.module) pkg.main = pkg.module;
    return pkg;
  }
});
const mappingsFile = normalizePath(path.dirname(vue3ui), "mappings.json");
const mappings = require(mappingsFile); // eslint-disable-line @typescript-eslint/no-var-requires
const unstablesFile = normalizePath(path.dirname(vue3ui), "classes.json");
const unstables = require(unstablesFile); // eslint-disable-line @typescript-eslint/no-var-requires
function analyze(input, debug, filter, alias) {
  const extensions = [".ts", ".tsx", ".mjs", ".js", ".jsx", ".vue"];
  const isSupported = id => {
    const lowerId = normalizePath(id.toLowerCase());
    if (!filter(lowerId)) return false;
    return extensions.some(ext => lowerId.endsWith(ext));
  };
  const traversed = new Set();
  const whitelist = new Set(["*", "html", "head", "body", "div", "app"]);
  let currentTag = "";
  const idList = (Array.isArray(input) ? input : typeof input === "object" ? Object.values(input) : [input]).map(id => normalizePath(path.resolve(id)));
  const parser = new htmlparser.Parser({
    // all tags will be whitelisted
    // additionally, if it happens to be a vue3-ui component, all classes that have no dependencies must be whitelisted as well
    onopentagname(name) {
      whitelist.add(name);
      currentTag = pascalCase(name);
      if (kebabCase(currentTag).startsWith("v-")) {
        var _unstables$currentTag;
        const props = (_unstables$currentTag = unstables[currentTag]) !== null && _unstables$currentTag !== void 0 ? _unstables$currentTag : {};
        const classes = Object.keys(props);
        for (const cl of classes) props[cl].length === 0 && whitelist.add(cl);
      }
    },
    onattribute(p, data) {
      for (const cl of data.split(" ")) whitelist.add(cl);
      const prop = p.replace(":", ""); // remove : from props like :loading -> loading
      if (kebabCase(currentTag).startsWith("v-")) {
        var _mappings$currentTag, _unstables$currentTag2;
        // optional
        if ((_mappings$currentTag = mappings[currentTag]) !== null && _mappings$currentTag !== void 0 && _mappings$currentTag.optional.includes(`is-${prop}`)) {
          console.log(`${currentTag} optional prop:`, `is-${prop}`);
          whitelist.add(`is-${prop}`);
          return;
        }
        // unstable
        const p = camelCase(prop);
        const props = (_unstables$currentTag2 = unstables[currentTag]) !== null && _unstables$currentTag2 !== void 0 ? _unstables$currentTag2 : {};
        const classes = Object.keys(props);
        for (const cl of classes) {
          const valid = props[cl].includes(p);
          valid && console.log(`${currentTag} unstable prop:`, p);
          valid && whitelist.add(cl);
        }
      }
    }
  }, {
    decodeEntities: true,
    lowerCaseTags: false,
    lowerCaseAttributeNames: true
  });
  function extract(code, id) {
    debug && console.log(`ANALYZER - TRAVERSAL (${humanlizePath(id)})`);
    if (!isVueSFC(id)) return code;
    debug && console.log(`ANALYZER - INCLUDED (${humanlizePath(id)})`);
    const {
      template,
      script
    } = parseSFC(code, id);
    if (template !== null && template !== void 0 && template.content) parser.parseComplete(template.content);
    if (script !== null && script !== void 0 && script.content) return script.content;
    return;
  }
  function resolveSource(id, value) {
    // Remove query
    value = value.split("?", 2)[0];
    // Resolve aliases
    for (const [from, to] of Object.entries(alias)) {
      if (value !== from && !value.startsWith(`${from}/`)) continue;
      value = to + value.slice(from.length);
    }
    const depId = normalizePath(sync(value, {
      basedir: path.dirname(id),
      extensions,
      packageFilter(pkg) {
        // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access, @typescript-eslint/no-unsafe-assignment
        if (pkg.module) pkg.main = pkg.module;
        // eslint-disable-next-line @typescript-eslint/no-unsafe-return
        return pkg;
      }
    }));
    if (traversed.has(depId) || !isSupported(depId)) return;else traversed.add(depId);
    return depId;
  }
  function traverseSource(id, rawCode) {
    const code = extract(rawCode, id);
    if (!code) return;
    const ast = jsparser.parse(code, parserOpts);
    traverse(ast, {
      // eslint-disable-next-line @typescript-eslint/naming-convention
      StringLiteral({
        node
      }) {
        if (!isVueSFC(id)) return;
        for (const cl of node.value.split(" ")) whitelist.add(cl);
      },
      // eslint-disable-next-line @typescript-eslint/naming-convention
      ExportNamedDeclaration({
        node
      }) {
        if (!node.source) return;
        const depId = resolveSource(id, node.source.value);
        if (depId) idList.push(depId);
      },
      // eslint-disable-next-line @typescript-eslint/naming-convention
      ImportDeclaration({
        node
      }) {
        const depId = resolveSource(id, node.source.value);
        if (depId) {
          idList.push(depId);
          if (!depId.includes("@pathscale/vue3-ui")) return;
        }
        for (const spec of node.specifiers) {
          var _mappings$spec$import;
          if (spec.type !== "ImportSpecifier") continue;
          if (!("name" in spec.imported)) continue;
          const wl = (_mappings$spec$import = mappings[spec.imported.name]) === null || _mappings$spec$import === void 0 ? void 0 : _mappings$spec$import.always;
          if (!wl) continue;
          for (const i of wl) whitelist.add(i);
          debug && console.log(`ANALYZER - VUE3-UI COMPONENT (${spec.imported.name}):\n`, wl);
        }
      }
    });
  }
  while (idList.length > 0) {
    const id = idList.pop();
    if (!id) continue;
    const code = fs.readFileSync(id, "utf8");
    try {
      traverseSource(id, code);
    } catch (error) {
      debug && console.log(`ANALYZER - PROCESSING ERROR (${humanlizePath(id)}):\n`, error);
    }
  }
  const wl = [...whitelist]
  // Delete some garbage
  .filter(_v => {
    const v = _v.trim();
    const garbage = ["vue", "slot"];
    if (!v) return false;
    if (/[A-Z]/.test(v)) return false;
    if (/[./:\\]/.test(v)) return false;
    if (garbage.includes(v)) return false;
    return true;
  }).sort();
  return wl;
}

var postCleaner = code => {
  let _code = code;
  const originalCount = getDeclaredVariables(_code).length;
  let purgedCount = 0;
  for (;;) {
    const vars = getDeclaredVariables(_code);
    const unused = vars.filter(v => !isVariableUsed(v, _code));
    if (unused.length === 0) break;
    for (const v of unused) {
      const decl = new RegExp(`.*--${v}:.*`, "g");
      _code = _code.replace(decl, "");
      if (_code.includes(`--${v}:`)) throw new Error("hey");
    }
    purgedCount += unused.length;
  }
  console.log(`POST-CLEANER - VARIABLES (${purgedCount} out of ${originalCount})`);
  return _code;
};

const generator = (options = {}) => {
  var _options$exclude;
  const filter = createFilter(options.include, (_options$exclude = options.exclude) !== null && _options$exclude !== void 0 ? _options$exclude : ["**/node_modules/**"]);
  const isVue3UICSS = createFilter(["**/node_modules/@pathscale/vue3-ui/**/*.css", "**/node_modules/@pathscale/bulma-css-var-only/**/*.css", "**/node_modules/@pathscale/bulma-extensions-css-var/**/*.css", "**/node_modules/@bulvar/bulma/css/bulma.css"]);
  let foundMain = false;
  let fakeBundlePath = "";
  let base = [];
  const whitelist = new Set();
  const plugin = {
    name: "vue3-ui-css-purge",
    buildStart(inputOpts) {
      var _options$debug, _options$alias;
      base = ["*", "html", "head", "body", "app", "div", ...analyze(inputOpts.input, (_options$debug = options.debug) !== null && _options$debug !== void 0 ? _options$debug : false, filter, (_options$alias = options.alias) !== null && _options$alias !== void 0 ? _options$alias : {})].map(b => b.replace(/[$()*+.?[\\\]^{|}-]/g, "\\$&"));
      for (const b of base) {
        whitelist.add(new RegExp(`^${b}$`));
        whitelist.add(new RegExp(`${b}\\[.+?\\]`));
      }
      options.debug && console.log(`CSS PURGER - WHITELIST:\n`, inspect(new Set([...base]), {
        showHidden: false,
        depth: null,
        maxArrayLength: null
      }));
    },
    async transform(code, id) {
      // if (id.includes(".vue?vue&type=template")) console.log(code);
      if (isVue3UICSS(id)) return "";
      if (isMain(code) && !foundMain) {
        const newJs = injectFakeBundle(code);
        const fakeBundle = await makeVue3UiBundle(id);
        foundMain = true;
        fakeBundlePath = `${path.dirname(id)}/vue3-ui-bundle.css`;
        await fs.writeFile(fakeBundlePath, fakeBundle);
        return newJs;
      }
      if (!id.includes("vue3-ui-bundle.css")) return null;
      /**
       * having as whitelist = ["switch", "input", "check"] will purge the following code anyway
       * .switch input[type=checkbox]+.check { ... }
       *
       * whitelisting all children for switch works, TODO: find the proper way to not nuke this
       * */
      const deepClasses = ["switch"];
      // eslint-disable-next-line unicorn/no-array-reduce
      const deep = deepClasses.reduce((acc, cl) => base.includes(cl) ? [...acc, new RegExp(cl)] : acc, []);
      const purger = postcss(purgecss({
        content: [],
        safelist: {
          standard: [...whitelist],
          deep
        },
        keyframes: true
      }));
      const {
        css
      } = await purger.process(code, {
        from: id
      });
      return {
        code: postCleaner(css)
      };
    },
    async buildEnd() {
      if (!(await fs.pathExists(fakeBundlePath))) this.warn("CSS PURGER - FAKE BUNDLE MISSING (vue3-ui-bundle.css)");
      await fs.remove(fakeBundlePath);
    }
  };
  return plugin;
};

export { generator as default };
